// https://github.com/meekrosoft/fff
#include <string.h>
#include <stdlib.h>
#include <setjmp.h>
#include <pthread.h>

// Want to keep the argument history for 512 calls
#define FFF_ARG_HISTORY_LEN 512
// Want to keep the call sequence history for 512 function calls
#define FFF_CALL_HISTORY_LEN 512

#include "fff/gtest/gtest.h"
#include "fff/fff.h"
DEFINE_FFF_GLOBALS;

extern pthread_mutex_t mutex; // テストケースを逐次実行するための保険

extern "C"
{
// テスト対象のCソースファイルを丸ごとインクルードすることにより、
// static関数やstatic変数の参照・更新が可能となる。
#include "../blink_sub.c"

	// グローバル変数の偽装
	TBL_CTL gbl_ctl;
}

/* 関数の偽装宣言 */
/* 関数の実体があるものはテスト対象のCソースファイルに記載の関数のみです。 */
/* 実体のない関数は、FFFで偽装します。 */
/* サービスコールの偽装宣言 */
/* 時間管理機能 */
FAKE_VALUE_FUNC(ER, sta_cyc, ID);
FAKE_VALUE_FUNC(ER, stp_cyc, ID);

/* ライブラリの偽装宣言 */
FAKE_VOID_FUNC(t_perror, uint_t, const char *, int_t, const char *, ER);

/* Device Driverの偽装宣言 */
FAKE_VOID_FUNC(set_led4_state, unsigned char);

/* List of fakes used by this unit tester */
#define FFF_FAKES_LIST(FAKE) \
	FAKE(sta_cyc)            \
	FAKE(stp_cyc)            \
	FAKE(set_led4_state)     \
	FAKE(t_perror)

/* テストケースの初期化と後処理 									*/
/* テスト対象、及びテストケースはグローバル変数を使用している。 	*/
/* マルチスレッド環境下で複数のテストケースが同時実行された場合、	*/
/* グローバル変数のアクセスでテストが破綻する。  					*/
/* 複数のテストケースを同時実行することは無いとは思うが、 			*/
/* 保険としてmutexにてテストケースを逐次実行するようにした。		*/
class testBlinkSub : public testing::Test
{
	void SetUp()
	{
		pthread_mutex_lock(&mutex);
		/* Register resets */
		FFF_FAKES_LIST(RESET_FAKE);
		/* reset common FFF internal structures */
		FFF_RESET_HISTORY();
	}
	void TearDown()
	{
		pthread_mutex_unlock(&mutex);
	}
};

// ブリンクタスクのサブルーチン(アクション関数)のテスト
// 状態:アイドル,イベント:タイムアウト受信時のアクション関数のテスト
TEST_F(testBlinkSub, testStIdleEvTimeOut)
{
	memset(&gbl_ctl, 0xFF, sizeof(TBL_CTL)); // 予めメモリを汚しておく
	gbl_ctl.ctl_flg = 0;

	StIdleEvTimeOut(); // テスト対象をコール

	// 期待値を確認
	EXPECT_EQ(DBL_STSCHG, gbl_ctl.ctl_flg);
	EXPECT_EQ(TIMEOUT_BLINK, gbl_ctl.bcount);
	EXPECT_EQ(ST_BLINK_LED_OFF, gbl_ctl.nextsts);
}

// 状態:アイドル,イベント:ACTIVE通知受信時のアクション関数のテスト
TEST_F(testBlinkSub, testStIdleEvAvtive)
{
	memset(&gbl_ctl, 0xFF, sizeof(TBL_CTL)); // 予めメモリを汚しておく
	gbl_ctl.ctl_flg = 0;

	StIdleEvAvtive(); // テスト対象をコール

	// 期待値を確認
	EXPECT_EQ(DBL_STSCHG, gbl_ctl.ctl_flg);
	EXPECT_EQ(ACT_BLINK_TIME, gbl_ctl.bcount);
	EXPECT_EQ(ST_BLINK_LED_OFF, gbl_ctl.nextsts);

	// sta_cycがコールされていることを確認
	EXPECT_EQ(fff.call_history[0], (void *)sta_cyc);
	// sta_cycの第1引数が正しいことを確認
	EXPECT_EQ(LED4_BLINK_TIME_HANDLER, sta_cyc_fake.arg0_history[0]);
}

// 状態:点滅状態(消灯),イベント:OFF通知受信時のアクション関数のテスト
TEST_F(testBlinkSub, testStLedOffEvOff)
{
	memset(&gbl_ctl, 0xFF, sizeof(TBL_CTL)); // 予めメモリを汚しておく
	gbl_ctl.ctl_flg = 0;

	StLedOffEvOff(); // テスト対象をコール

	// 期待値を確認
	EXPECT_EQ(DBL_STSCHG, gbl_ctl.ctl_flg);
	EXPECT_EQ(0, gbl_ctl.bcount);
	EXPECT_EQ(ST_BLINK_IDLE, gbl_ctl.nextsts);

	// set_led4_stateがコールされていることを確認
	EXPECT_EQ(fff.call_history[0], (void *)set_led4_state);
	// set_led4_stateの第1引数が正しいことを確認
	EXPECT_EQ(OFF, set_led4_state_fake.arg0_history[0]);

	// stp_cycがコールされていることを確認
	EXPECT_EQ(fff.call_history[1], (void *)stp_cyc);
	// stp_cycの第1引数が正しいことを確認
	EXPECT_EQ(LED4_BLINK_TIME_HANDLER, stp_cyc_fake.arg0_history[0]);
}

// 状態:点滅状態(点灯),イベント:OFF通知受信時のアクション関数のテスト
TEST_F(testBlinkSub, testStLedOnEvOff)
{
	memset(&gbl_ctl, 0xFF, sizeof(TBL_CTL)); // 予めメモリを汚しておく
	gbl_ctl.ctl_flg = 0;

	StLedOnEvOff(); // テスト対象をコール

	// 期待値を確認
	EXPECT_EQ(DBL_STSCHG, gbl_ctl.ctl_flg);
	EXPECT_EQ(0, gbl_ctl.bcount);
	EXPECT_EQ(ST_BLINK_IDLE, gbl_ctl.nextsts);

	// set_led4_stateがコールされていることを確認
	EXPECT_EQ(fff.call_history[0], (void *)set_led4_state);
	// set_led4_stateの第1引数が正しいことを確認
	EXPECT_EQ(OFF, set_led4_state_fake.arg0_history[0]);

	// stp_cycがコールされていることを確認
	EXPECT_EQ(fff.call_history[1], (void *)stp_cyc);
	// stp_cycの第1引数が正しいことを確認
	EXPECT_EQ(LED4_BLINK_TIME_HANDLER, stp_cyc_fake.arg0_history[0]);
}

// 状態:点滅状態(消灯),イベント:0.25秒経過受信のアクション関数のテスト(点滅回数ラスト1回の場合)
TEST_F(testBlinkSub, testStLedOffEvBlinkLast)
{
	memset(&gbl_ctl, 0xFF, sizeof(TBL_CTL)); // 予めメモリを汚しておく
	gbl_ctl.ctl_flg = 0;

	GBL_BCOUNT = 1;	   // 残り点滅回数1回
	StLedOffEvBlink(); // テスト対象をコール

	// 期待値を確認
	EXPECT_EQ(DBL_STSCHG, gbl_ctl.ctl_flg);
	EXPECT_EQ(0, gbl_ctl.bcount);
	EXPECT_EQ(ST_BLINK_IDLE, gbl_ctl.nextsts);

	// stp_cycがコールされていることを確認
	EXPECT_EQ(fff.call_history[0], (void *)stp_cyc);
	// stp_cycの第1引数が正しいことを確認
	EXPECT_EQ(LED4_BLINK_TIME_HANDLER, stp_cyc_fake.arg0_history[0]);
}

// 状態:点滅状態(消灯),イベント:0.25秒経過受信のアクション関数のテスト(点滅継続の場合)
TEST_F(testBlinkSub, testStLedOffEvBlinkContinue)
{
	memset(&gbl_ctl, 0xFF, sizeof(TBL_CTL)); // 予めメモリを汚しておく
	gbl_ctl.ctl_flg = 0;

	GBL_BCOUNT = ACT_BLINK_TIME; // 残り点滅回数4回
	StLedOffEvBlink();			 // テスト対象をコール

	// 期待値を確認
	EXPECT_EQ(DBL_STSCHG, gbl_ctl.ctl_flg);
	EXPECT_EQ((ACT_BLINK_TIME - 1), gbl_ctl.bcount);
	EXPECT_EQ(ST_BLINK_LED_ON, gbl_ctl.nextsts);

	// set_led4_stateがコールされていることを確認
	EXPECT_EQ(fff.call_history[0], (void *)set_led4_state);
	// set_led4_stateの第1引数が正しいことを確認
	EXPECT_EQ(ON, set_led4_state_fake.arg0_history[0]);
}

// 状態:点滅状態(点灯),イベント:0.25秒経過受信のアクション関数のテスト(点滅回数ラスト1回の場合)
TEST_F(testBlinkSub, testStLedOnEvBlinkLast)
{
	memset(&gbl_ctl, 0xFF, sizeof(TBL_CTL)); // 予めメモリを汚しておく
	gbl_ctl.ctl_flg = 0;

	GBL_BCOUNT = 1;	  // 残り点滅回数1回
	StLedOnEvBlink(); // テスト対象をコール

	// 期待値を確認
	EXPECT_EQ(DBL_STSCHG, gbl_ctl.ctl_flg);
	EXPECT_EQ(0, gbl_ctl.bcount);
	EXPECT_EQ(ST_BLINK_IDLE, gbl_ctl.nextsts);

	// set_led4_stateがコールされていることを確認
	EXPECT_EQ(fff.call_history[0], (void *)set_led4_state);
	// set_led4_stateの第1引数が正しいことを確認
	EXPECT_EQ(OFF, set_led4_state_fake.arg0_history[0]);

	// stp_cycがコールされていることを確認
	EXPECT_EQ(fff.call_history[1], (void *)stp_cyc);
	// stp_cycの第1引数が正しいことを確認
	EXPECT_EQ(LED4_BLINK_TIME_HANDLER, stp_cyc_fake.arg0_history[0]);
}

// 状態:点滅状態(点灯),イベント:0.25秒経過受信のアクション関数のテスト(点滅継続の場合)
TEST_F(testBlinkSub, testStLedOnEvBlinkContinue)
{
	memset(&gbl_ctl, 0xFF, sizeof(TBL_CTL)); // 予めメモリを汚しておく
	gbl_ctl.ctl_flg = 0;

	GBL_BCOUNT = ACT_BLINK_TIME; // 残り点滅回数4回
	StLedOnEvBlink();			 // テスト対象をコール

	// 期待値を確認
	EXPECT_EQ(DBL_STSCHG, gbl_ctl.ctl_flg);
	EXPECT_EQ((ACT_BLINK_TIME - 1), gbl_ctl.bcount);
	EXPECT_EQ(ST_BLINK_LED_OFF, gbl_ctl.nextsts);

	// set_led4_stateがコールされていることを確認
	EXPECT_EQ(fff.call_history[0], (void *)set_led4_state);
	// set_led4_stateの第1引数が正しいことを確認
	EXPECT_EQ(OFF, set_led4_state_fake.arg0_history[0]);
}
