// https://github.com/meekrosoft/fff
#include <string.h>
#include <stdlib.h>
#include <setjmp.h>
#include <pthread.h>

// Want to keep the argument history for 512 calls
#define FFF_ARG_HISTORY_LEN 512
// Want to keep the call sequence history for 512 function calls
#define FFF_CALL_HISTORY_LEN 512

#include "fff/gtest/gtest.h"
#include "fff/fff.h"
DEFINE_FFF_GLOBALS;

extern pthread_mutex_t mutex; // テストケースを逐次実行するための保険

extern "C"
{
// テスト対象のCソースファイルを丸ごとインクルードすることにより、
// static関数やstatic変数の参照・更新が可能となる。
#include "../timer_sub.c"

    // グローバル変数の偽装
    TTM_CTL gtm_ctl;
}

/* 関数の偽装宣言 */
/* 関数の実体があるものはテスト対象のCソースファイルに記載の関数のみです。 */
/* 実体のない関数は、FFFで偽装します。 */

/* サービスコールの偽装宣言 */

/* 時間管理機能 */
FAKE_VALUE_FUNC(ER, sta_cyc, ID);
FAKE_VALUE_FUNC(ER, stp_cyc, ID);
FAKE_VALUE_FUNC(ER, get_utm, SYSUTM *)

/* 同期・通信機能 */
FAKE_VALUE_FUNC(ER, snd_dtq, ID, intptr_t);

/* ライブラリの偽装宣言 */
FAKE_VOID_FUNC(t_perror, uint_t, const char *, int_t, const char *, ER);

// // 偽装データキュー送信関数（結果のみを返す）→ 必要なし
// ER snd_dtq_custom_fake(ID dtqid, intptr_t data)
// {
//     return (E_OK);
// }

/* List of fakes used by this unit tester */
#define FFF_FAKES_LIST(FAKE) \
    FAKE(sta_cyc)            \
    FAKE(stp_cyc)            \
    FAKE(get_utm)            \
    FAKE(snd_dtq)            \
    FAKE(t_perror)

/* テストケースの初期化と後処理 									*/
/* テスト対象、及びテストケースはグローバル変数を使用している。 	*/
/* マルチスレッド環境下で複数のテストケースが同時実行された場合、	*/
/* グローバル変数のアクセスでテストが破綻する。  					*/
/* 複数のテストケースを同時実行することは無いとは思うが、 			*/
/* 保険としてmutexにてテストケースを逐次実行するようにした。		*/
class testTimerSub : public testing::Test
{
    void SetUp()
    {
        pthread_mutex_lock(&mutex);
        /* Register resets */
        FFF_FAKES_LIST(RESET_FAKE);
        /* reset common FFF internal structures */
        FFF_RESET_HISTORY();
    }
    void TearDown()
    {
        pthread_mutex_unlock(&mutex);
    }
};

// タイマー開始処理関数の検証
TEST_F(testTimerSub, test_StOnEntry)
{
    // メモリを汚す
    memset(&gtm_ctl, 0xFF, sizeof(gtm_ctl));

    // 処理前制御オブジェクトのバックアップ（検証用）
    TTM_CTL gtm_ctl_init = gtm_ctl;

    // テスト対象の呼び出し
    StOnEntry();

    // タイムアウト時間設定値の検証
    EXPECT_EQ(INIT_TIME, gtm_ctl.timer);

    // スタート時間記録関数の呼び出し検証
    EXPECT_EQ(1, get_utm_fake.call_count);

    // タイマー開始関数の呼び出し検証
    EXPECT_EQ(1, sta_cyc_fake.call_count);
    EXPECT_EQ(BASE_TIME_HANDLER, sta_cyc_fake.arg0_val);

    // その他値が変更されていないことの検証
    EXPECT_EQ(gtm_ctl_init.ctl_flg, gtm_ctl.ctl_flg);
    EXPECT_EQ(gtm_ctl_init.nextsts, gtm_ctl.nextsts);
    EXPECT_EQ(gtm_ctl_init.nowsts, gtm_ctl.nowsts);
    EXPECT_EQ(gtm_ctl_init.prcv_data, gtm_ctl.prcv_data);
}

// タイマーOFF中 SW1状態更新時（OFF）処理関数の検証
TEST_F(testTimerSub, test_StOffEvSw1Off)
{
    // データキュー送信関数の偽装（必要なかった）
    // snd_dtq_fake.custom_fake = snd_dtq_custom_fake;

    // メモリを汚す
    memset(&gtm_ctl, 0xFF, sizeof(gtm_ctl));
    // データキュー受信偽装（受信先アドレスの取得のみ）
    gtm_ctl.prcv_data = (T_EVENT *)malloc(sizeof(T_EVENT));

    // 処理前制御オブジェクトのバックアップ（検証用）
    TTM_CTL gtm_ctl_init = gtm_ctl;

    // テスト対象の呼び出し
    StOffEvSw1Off();

    // セットされるべき値がセットされていることの検証
    // コントロールフラグの検証
    EXPECT_EQ(gtm_ctl_init.ctl_flg | DTM_NOTREL, gtm_ctl.ctl_flg);
    // メッセージの検証
    EXPECT_EQ(EV_BLINK_OFF, gtm_ctl.prcv_data->event);

    // データキュー送信の検証（関数呼び出し、送信先、送信内容の検証）
    EXPECT_EQ(1, snd_dtq_fake.call_count);
    EXPECT_EQ(BLINK_DTQ, snd_dtq_fake.arg0_val);
    EXPECT_EQ((intptr_t)gtm_ctl.prcv_data, snd_dtq_fake.arg1_val);

    // その他値が変更されていないことの検証
    EXPECT_EQ(gtm_ctl_init.nextsts, gtm_ctl.nextsts);
    EXPECT_EQ(gtm_ctl_init.nowsts, gtm_ctl.nowsts);
    EXPECT_EQ(gtm_ctl_init.prcv_data, gtm_ctl.prcv_data);
    EXPECT_EQ(gtm_ctl_init.timer, gtm_ctl.timer);

    free(gtm_ctl.prcv_data);
}

// タイマーON中 SW1状態更新時（OFF）処理関数の検証
TEST_F(testTimerSub, test_StOnEvSw1Off)
{
    // メモリを汚す
    memset(&gtm_ctl, 0xFF, sizeof(gtm_ctl));
    // データキュー受信偽装（受信先アドレスの取得のみ）
    gtm_ctl.prcv_data = (T_EVENT *)malloc(sizeof(T_EVENT));

    // 処理前コントロールフラグのバックアップ（検証用）
    uint32_t ctl_flg_init = gtm_ctl.ctl_flg;

    // テスト対象の呼び出し
    StOnEvSw1Off();

    // コントロールフラグの変更処理検証
    EXPECT_EQ(ctl_flg_init | (DTM_STSCHG | DTM_NOTREL), gtm_ctl.ctl_flg);

    // 次状態フラグの変更処理検証
    EXPECT_EQ(ST_TIMER_OFF, gtm_ctl.nextsts);

    // データキュー送信の検証（関数呼び出し、送信先、送信内容の検証）
    EXPECT_EQ(1, snd_dtq_fake.call_count);
    EXPECT_EQ(BLINK_DTQ, snd_dtq_fake.arg0_val);
    EXPECT_EQ((intptr_t)gtm_ctl.prcv_data, snd_dtq_fake.arg1_val);
    EXPECT_EQ(EV_BLINK_OFF, gtm_ctl.prcv_data->event);

    free(gtm_ctl.prcv_data);
}

// タイマーOFF中 SW1状態更新時（ON）処理関数の検証
TEST_F(testTimerSub, test_StOffEvSw1On)
{
    // メモリを汚す
    memset(&gtm_ctl, 0xFF, sizeof(gtm_ctl));
    // データキュー受信偽装（受信先アドレスの取得のみ）
    gtm_ctl.prcv_data = (T_EVENT *)malloc(sizeof(T_EVENT));

    // 処理前コントロールフラグのバックアップ（検証用）
    uint32_t ctl_flg_init = gtm_ctl.ctl_flg;

    // テスト対象の呼び出し
    StOffEvSw1On();

    // コントロールフラグの変更処理検証
    EXPECT_EQ(ctl_flg_init | (DTM_STSCHG | DTM_NOTREL), gtm_ctl.ctl_flg);

    // 次状態フラグの変更処理検証
    EXPECT_EQ(ST_TIMER_ON, gtm_ctl.nextsts);

    // データキュー送信の検証（関数呼び出し、送信先、送信内容の検証）
    EXPECT_EQ(1, snd_dtq_fake.call_count);
    EXPECT_EQ(BLINK_DTQ, snd_dtq_fake.arg0_val);
    EXPECT_EQ((intptr_t)gtm_ctl.prcv_data, snd_dtq_fake.arg1_val);
    EXPECT_EQ(EV_BLINK_ACTIVE, gtm_ctl.prcv_data->event);

    free(gtm_ctl.prcv_data);
}

// タイマーON中 Push1状態更新時（ON）処理関数の検証
TEST_F(testTimerSub, test_StOnEvPush1On)
{
    // メモリを汚す
    memset(&gtm_ctl, 0xFF, sizeof(gtm_ctl));

    // 処理前タイムアウト値のバックアップ（検証用）
    int timer_init = gtm_ctl.timer;

    // テスト対象の呼び出し
    StOnEvPush1On();

    // タイムアウト値の変更処理検証
    EXPECT_EQ(timer_init + EXTRA_UNIT, gtm_ctl.timer);
}

// 定期（1秒毎）処理関数の検証 - CASE：Non-Zero and Non-MulOf10
TEST_F(testTimerSub, test_NonZero_StOnEvBaseTime)
{
    // タイマーカウント値をセット（非10の倍数かつ非0）
    GTM_REMTM = 7;

    // メモリを汚す
    memset(&gtm_ctl, 0xFF, sizeof(gtm_ctl));

    // 処理前制御オブジェクトのバックアップ（検証用）
    TTM_CTL gtm_ctl_init = gtm_ctl;

    // テスト対象の呼び出し
    StOnEvBaseTime();

    // タイムアウト値が -1 されていることの検証
    EXPECT_EQ(gtm_ctl_init.timer - 1, gtm_ctl.timer);

    // その他値が変更されていないことの検証
    EXPECT_EQ(gtm_ctl_init.ctl_flg, gtm_ctl.ctl_flg);
    EXPECT_EQ(gtm_ctl_init.nextsts, gtm_ctl.nextsts);
    EXPECT_EQ(gtm_ctl_init.nowsts, gtm_ctl.nowsts);
    EXPECT_EQ(gtm_ctl_init.prcv_data, gtm_ctl.prcv_data);
}

// 定期（1秒毎）処理関数の検証 - CASE：LastCall
TEST_F(testTimerSub, test_LastCall_StOnEvBaseTime)
{
    // メモリを汚す
    memset(&gtm_ctl, 0xFF, sizeof(gtm_ctl));
    // データキュー受信偽装（受信先アドレスの取得のみ）
    gtm_ctl.prcv_data = (T_EVENT *)malloc(sizeof(T_EVENT));

    // タイマーカウント値をセット（タイムアウト値が1の場合に最終呼び出しとなる）
    GTM_REMTM = 1;

    // 処理前制御オブジェクトのバックアップ（検証用）
    TTM_CTL gtm_ctl_init = gtm_ctl;

    // テスト対象の呼び出し
    StOnEvBaseTime();

    // タイムアウト値が -1 されていることの検証
    EXPECT_EQ(gtm_ctl_init.timer - 1, gtm_ctl.timer);

    // 呼ばれるべき関数が呼ばれていることの検証
    EXPECT_EQ(1, get_utm_fake.call_count);
    EXPECT_EQ(1, snd_dtq_fake.call_count);

    // セットされるべき値がセットされていることの検証
    EXPECT_EQ(gtm_ctl_init.ctl_flg | (DTM_STSCHG | DTM_NOTREL), gtm_ctl.ctl_flg);
    EXPECT_EQ(ST_TIMER_OFF, gtm_ctl.nextsts);
    EXPECT_EQ(EV_BLINK_TIMEOUT, gtm_ctl.prcv_data->event);

    // その他値が不意に変更されていないことの検証
    EXPECT_EQ(gtm_ctl_init.nowsts, gtm_ctl.nowsts);
    EXPECT_EQ(gtm_ctl_init.prcv_data, gtm_ctl.prcv_data);

    // データキュー送信内容の検証（送信先、送信内容）
    EXPECT_EQ(BLINK_DTQ, snd_dtq_fake.arg0_val);
    EXPECT_EQ((intptr_t)gtm_ctl.prcv_data, snd_dtq_fake.arg1_val);

    free(gtm_ctl.prcv_data);
}

// 定期（1秒毎）処理関数の検証 - CASE：MultipleOf10
TEST_F(testTimerSub, test_MulOf10_StOnEvBaseTime)
{
    // メモリを汚す
    memset(&gtm_ctl, 0xFF, sizeof(gtm_ctl));
    // データキュー受信偽装（受信先アドレスの取得のみ）
    gtm_ctl.prcv_data = (T_EVENT *)malloc(sizeof(T_EVENT));

    // タイマーカウント値をセット（タイムアウト値が1の場合に最終呼び出しとなる）
    GTM_REMTM = 11;

    // 処理前制御オブジェクトのバックアップ（検証用）
    TTM_CTL gtm_ctl_init = gtm_ctl;

    // テスト対象の呼び出し
    StOnEvBaseTime();

    // タイムアウト値が -1 されていることの検証
    EXPECT_EQ(gtm_ctl_init.timer - 1, gtm_ctl.timer);

    // 呼ばれるべき関数が呼ばれていることの検証
    EXPECT_EQ(1, get_utm_fake.call_count);
    EXPECT_EQ(1, snd_dtq_fake.call_count);

    // セットされるべき値がセットされていることの検証
    EXPECT_EQ(gtm_ctl_init.ctl_flg | DTM_NOTREL, gtm_ctl.ctl_flg);
    EXPECT_EQ(EV_BLINK_ACTIVE, gtm_ctl.prcv_data->event);

    // その他値が不意に変更されていないことの検証
    EXPECT_EQ(gtm_ctl_init.ctl_flg, gtm_ctl.ctl_flg);
    EXPECT_EQ(gtm_ctl_init.nowsts, gtm_ctl.nowsts);
    EXPECT_EQ(gtm_ctl_init.nextsts, gtm_ctl.nextsts);
    EXPECT_EQ(gtm_ctl_init.prcv_data, gtm_ctl.prcv_data);

    // データキュー送信内容の検証（送信先、送信内容）
    EXPECT_EQ(BLINK_DTQ, snd_dtq_fake.arg0_val);
    EXPECT_EQ((intptr_t)gtm_ctl.prcv_data, snd_dtq_fake.arg1_val);

    free(gtm_ctl.prcv_data);
}

// タイマー終了時処理関数の検証
TEST_F(testTimerSub, test_StOnExit)
{
    // メモリを汚す
    memset(&gtm_ctl, 0xFF, sizeof(gtm_ctl));

    // 処理前制御オブジェクトのバックアップ（検証用）
    TTM_CTL gtm_ctl_init = gtm_ctl;

    // テスト対象の呼び出し
    StOnExit();

    // セットされるべき値がセットされていることの検証
    EXPECT_EQ(0, gtm_ctl.timer);

    // 呼ばれるべき関数が呼ばれていることの検証
    EXPECT_EQ(1, stp_cyc_fake.call_count);

    // その他値が不意に変更されていないことの検証
    EXPECT_EQ(gtm_ctl_init.ctl_flg, gtm_ctl.ctl_flg);
    EXPECT_EQ(gtm_ctl_init.nextsts, gtm_ctl.nextsts);
    EXPECT_EQ(gtm_ctl_init.nowsts, gtm_ctl.nowsts);
    EXPECT_EQ(gtm_ctl_init.prcv_data, gtm_ctl.prcv_data);
}